@@ -26,6 +26,8 @@ import com.vizdom.ber.BerTypes;
 import com.vizdom.util.Debug;
 import com.vizdom.util.UnreachableCodeException;
 import java.io.*;
+import java.lang.reflect.InvocationTargetException;
+import java.lang.reflect.Method;
 import java.math.BigDecimal;
 import java.net.Socket;
 import java.sql.*;
@@ -37,7 +39,7 @@ import java.util.Hashtable;
  * connection-accepting server.
  *
  * @author Gennis Emerson
- * @version $Revision: 24 $
+ * @version $Revision: 1.26 $
  */
 /* The handleMessage methods may throw only SQLExceptions and 
  * RuntimeExceptions, the first if they can recover, the second
@@ -156,13 +158,13 @@ class Connection implements Runnable
                  */
                 request = mBerModule.readFrom(mIn);
                 if (request == null)
-                {
+                 {
                     throw new FatalException("Client disconnected");
                 }
                 Debug.getLogWriter(Debug.BRIEF).println(mThreadId + 
                     "request: " + request);
 
-                BerIdentifier id = request.getIdentifier();
+                BerIdentifier id = request.getIdentifier(); 
                 Debug.assert(id.getTagClass() == BerTypes.APPLICATION);
                 int tagNumber = id.getTagNumber();
                 switch (tagNumber)
@@ -226,6 +228,17 @@ class Connection implements Runnable
                     response = handleRequest(
                         (StatementDestroyRequest) request); 
                     break;
+
+                case BerDbdModule.gCONNECTION_FUNC_REQUEST: 
+                    response = handleRequest(
+                        (ConnectionFuncRequest) request); 
+                    break;
+
+                case BerDbdModule.gSTATEMENT_FUNC_REQUEST: 
+                    response = handleRequest(
+                        (StatementFuncRequest) request); 
+                    break;
+
                 default: 
                     throw new DbdException(DbdException.gUNKNOWN_REQUEST,
                         new String[] { String.valueOf(tagNumber) });
@@ -635,7 +648,7 @@ class Connection implements Runnable
         StatementHolder holder = mGetStatementHolder(aRequest.getHandle());
         ResultSet rs = holder.getResultSet();
         if (rs == null)
-            throw new DbdException(DbdException.gNO_DATA);
+            throw new DbdException(DbdException.gNO_RESULT_SET);
         
         Object[] row = null;
         boolean hasData;
@@ -980,24 +993,392 @@ class Connection implements Runnable
         // ??? Should this be a runtime exception? Can the user cause this?
         if (holder == null)
             throw new DbdException(DbdException.gINVALID_STATEMENT_HANDLE);
+        else
+        {
+            Debug.getLogWriter(Debug.VERBOSE).println(mThreadId +
+                "destroying statement " + handle);
+            
+            // Close the statement and result set. This may cause
+            // a commit or rollback in the JDBC driver in
+            // AutoCommit mode, but not closing these objects
+            // seems to be worse.
+            holder.close();
+            holder = null;
+        }
         return new StatementDestroyResponse();
     }    
 
 
+
+    /**
+     * Uses reflection to call a method on the current Connection object.
+     *
+     * @param aRequest contains the method name and parameter values
+     * @return a response packet containing the return value from the 
+     *      called method
+     * @exception SQLException if the invoked method throws a SQLException
+     * @exception DbdException if an error occurs
+     */
+    BerObject handleRequest(ConnectionFuncRequest aRequest)
+        throws SQLException,DbdException
+    {
+        String methodName = aRequest.getMethodName();
+        Debug.getLogWriter(Debug.VERBOSE).println(mThreadId +
+            "func method: " + methodName);
+        String value =  mHandleFunc(mConn, aRequest.getMethodName(), 
+            aRequest.getParameters());
+        Debug.getLogWriter(Debug.VERBOSE).println(mThreadId +
+            methodName + " returned '" + value + "'");
+        try
+        {
+            return new ConnectionFuncResponse(value, 
+                mBerModule.getCharacterEncoding());
+        }
+        catch (UnsupportedEncodingException unsupEnc)
+        { 
+            throw new DbdException(DbdException.gUNSUPPORTED_ENCODING,
+                new String[] { mBerModule.getCharacterEncoding() });
+        }
+    }
+
+    /**
+     * Uses reflection to call a method on a Statement,
+     * ResultSet, or ResultSetMetaData object. The particular
+     * statement is indicated by a statement handle in the
+     * request. The object is indicated by a
+     * <code>Statement.</code>, <code>ResultSet.</code>, or
+     * <code>ResultSetMetaData.</code> prefix on the method name.
+     *
+     * @param aRequest contains a statement handle, 
+     *      the method name and parameter values
+     * @return a response packet containing the return value from the 
+     *      called method 
+     * @exception SQLException if the invoked method throws a SQLException
+     * @exception DbdException if an error occurs
+     */
+    BerObject handleRequest(StatementFuncRequest aRequest)
+        throws SQLException,DbdException
+    {
+        String methodName = aRequest.getMethodName();
+        Debug.getLogWriter(Debug.VERBOSE).println(mThreadId +
+            "func method: " +  methodName);
+
+        int dot = methodName.lastIndexOf(".");
+        if (dot == -1)
+            throw new DbdException(DbdException.gREFLECTION_OBJECT_MISSING);
+
+        StatementHolder holder = mGetStatementHolder(aRequest.getHandle());
+        if (holder == null)
+            throw new DbdException(DbdException.gINVALID_STATEMENT_HANDLE);
+            
+        String objectName = methodName.substring(0, dot);
+        Debug.getLogWriter(Debug.TEDIOUS).println(mThreadId + 
+            "looking up object type " + objectName);
+        Object object = null;
+        if (objectName.equals("Statement") || 
+            objectName.equals("PreparedStatement"))
+        {
+            object = holder.getStatement();
+        }
+        else if (objectName.equals("ResultSet"))
+        {
+            object = holder.getResultSet();
+            if (object == null)
+                throw new DbdException(DbdException.gNO_RESULT_SET);
+        }
+        else if (objectName.equals("ResultSetMetaData"))
+        {
+            object = holder.getResultSetMetaData();
+            if (object == null)
+                throw new DbdException(DbdException.gNO_METADATA);
+        }
+        else
+        {
+            throw new DbdException(DbdException.gREFLECTION_INVALID_OBJECT,
+                new String[] { objectName } );
+        }
+
+        Debug.getLogWriter(Debug.TEDIOUS).println(mThreadId + 
+            "underlying object type: " + object.getClass().getName());
+
+        String value = mHandleFunc(object, methodName.substring(dot + 1),
+            aRequest.getParameters());
+        Debug.getLogWriter(Debug.VERBOSE).println(mThreadId +
+            methodName + " returned '" + value + "'");
+        try
+        {
+            return new StatementFuncResponse(value, 
+                mBerModule.getCharacterEncoding());
+        }
+        catch (UnsupportedEncodingException unsupEnc)
+        { 
+            throw new DbdException(DbdException.gUNSUPPORTED_ENCODING,
+                new String[] { mBerModule.getCharacterEncoding() });
+        }
+    }
+
+
+    /**
+     * Uses reflection to call the given method on the given
+     * object with the given parameter list.
+     *
+     * @param anObject the object on which to call the method
+     * @param aMethodName the method to be called
+     * @param aParameterList a list of parameters to be passed to the method;
+     *      may not be null, but may have length 0
+     * @return the value of calling <code>toString</code> on the
+     *      Object returned from the method call
+     * @exception SQLException if the invoked method throws a SQLException
+     * @exception DbdException if an error occurs 
+     */
+    private String mHandleFunc(Object anObject, String aMethodName,  
+        Parameter[] aParameterList) throws SQLException,DbdException  
+    { 
+        Debug.getLogWriter(Debug.TEDIOUS).println(mThreadId +  
+            "method has " + aParameterList.length + " parameters"); 
+
+        Object[] parameterObjects = new Object[aParameterList.length];
+        Class[] parameterClasses = new Class[aParameterList.length];
+        try
+        {
+            for (int i = 0; i < aParameterList.length; i++)
+            {
+                if (aParameterList[i].value == null)
+                {
+                    parameterObjects[i] = null;
+                    parameterClasses[i] = mGetClass(aParameterList[i].type);
+                    Debug.getLogWriter(Debug.TEDIOUS).println(mThreadId + 
+                        "parameter " + (i + 1) + ": null; class " + 
+                        parameterClasses[i].getName());
+                    continue;
+                }
+
+                Debug.getLogWriter(Debug.TEDIOUS).print(mThreadId + 
+                    "parameter " + (i + 1) + ": ");
+                switch (aParameterList[i].type) 
+                {
+                case Types.BINARY:
+                case Types.VARBINARY:
+                case Types.LONGVARBINARY:
+                    parameterObjects[i] = 
+                        aParameterList[i].value.toByteArray();
+                    break;
+                case Types.TINYINT:
+                    parameterObjects[i] = 
+                        new Byte(aParameterList[i].value.toString());
+                    break;
+                case Types.SMALLINT:
+                    parameterObjects[i] = 
+                        new Short(aParameterList[i].value.toString());
+                    break;
+                case Types.INTEGER:
+                    parameterObjects[i] = 
+                        new Integer(aParameterList[i].value.toString());
+                    break;
+                case Types.BIGINT: 
+                    parameterObjects[i] = 
+                        new Long(aParameterList[i].value.toString());
+                    break;
+                case Types.REAL: 
+                    parameterObjects[i] = 
+                        new Float(aParameterList[i].value.toString());
+                    break;
+                case Types.FLOAT: 
+                case Types.DOUBLE: 
+                    parameterObjects[i] = 
+                        new Double(aParameterList[i].value.toString());
+                    break;
+                case Types.DECIMAL: 
+                case Types.NUMERIC:
+                    parameterObjects[i] = 
+                        new BigDecimal(aParameterList[i].value.toString());
+                    break;
+                case Types.BIT:   // Clients must send "0" or "1"
+                    parameterObjects[i] = new Boolean(
+                        aParameterList[i].value.toString().equals("1"));
+                    break;
+                case Types.CHAR: 
+                case Types.VARCHAR:
+                case Types.LONGVARCHAR:  // Use a stream here?
+                    parameterObjects[i] = aParameterList[i].value.toString();
+                    break;
+                case Types.DATE:      
+                    parameterObjects[i] = 
+                        java.sql.Date.valueOf(aParameterList[i].value.toString());
+                    break;
+                case Types.TIME: 
+                    parameterObjects[i] = 
+                        java.sql.Time.valueOf(aParameterList[i].value.toString());
+                    break;
+                case Types.TIMESTAMP: 
+                    parameterObjects[i] = 
+                        java.sql.Timestamp.valueOf(aParameterList[i].value.toString());
+                    break;
+                case Types.OTHER: 
+                default: 
+                    parameterObjects[i] = aParameterList[i].value.toString();
+                    break;
+                }
+
+                parameterClasses[i] = mGetClass(aParameterList[i].type);
+
+                if (aParameterList[i].type == Types.BINARY ||
+                    aParameterList[i].type == Types.VARBINARY ||
+                    aParameterList[i].type == Types.LONGVARBINARY)
+                {
+                    Debug.getLogWriter(Debug.TEDIOUS).println(
+                        "byte[], length " + 
+                        ((byte[]) parameterObjects[i]).length +
+                        "; class " + parameterClasses[i].getName());
+                }
+                else
+                {
+                    Debug.getLogWriter(Debug.TEDIOUS).println(
+                        parameterObjects[i].toString() + "; class " + 
+                        parameterClasses[i].getName());
+                }
+            }
+            
+            try 
+            {
+                // Look up the method and call it.
+                Class c = anObject.getClass();
+                if (aParameterList.length == 0)
+                    parameterClasses = null;
+                Method method = c.getMethod(aMethodName, parameterClasses);
+                
+                Debug.getLogWriter(Debug.TEDIOUS).println(mThreadId + 
+                    "found method: " + method.toString());
+                
+                Object returnValue = null;
+                returnValue = method.invoke(anObject, parameterObjects ); 
+                // Null should be turned into BerNull. Boolean is
+                // turned into 1 or 0 so that the return value
+                // will be appropriately true or false in Perl.
+                if (returnValue == null)
+                {
+                    //System.out.println("returning null");
+                    return null;
+                }
+                else if (returnValue instanceof Boolean)
+                    return ((Boolean) returnValue).booleanValue() ? "1" : "0";
+                else
+                    return returnValue.toString();
+            }
+            catch (NoSuchMethodException noMethod)
+            {
+                throw new DbdException(DbdException.gREFLECTION_EXCEPTION,
+                    new String[] { noMethod.toString() });
+            }
+            catch (SecurityException security)
+            {
+                throw new DbdException(DbdException.gREFLECTION_EXCEPTION,
+                    new String[] { security.toString() });
+            }
+            catch (IllegalAccessException access)
+            {
+                throw new DbdException(DbdException.gREFLECTION_EXCEPTION,
+                    new String[] { access.toString() });
+            }
+            catch (IllegalArgumentException argument)
+            {
+                throw new DbdException(DbdException.gREFLECTION_EXCEPTION,
+                    new String[] { argument.toString() });
+            }
+            catch (InvocationTargetException target)
+            {
+                Throwable t = target.getTargetException();
+                if (t != null)
+                {
+                    if (t instanceof SQLException)
+                        throw (SQLException) t;
+                    throw new DbdException(DbdException.gREFLECTION_EXCEPTION,
+                        new String[] { t.toString() });
+                }
+                throw new DbdException(DbdException.gREFLECTION_EXCEPTION,
+                    new String[] { target.toString() });
+            }
+        }
+        catch (Exception e) 
+        { 
+            throw new DbdException(DbdException.gGENERIC_EXCEPTION,
+                new String[] { e.toString() });
+        }            
+    }
+
+
+    // Support methods.
+
+
+    /** 
+     * Returns a <code>java.lang.Class</code> object
+     * corresponding to the argument, which should be a value
+     * from <code>java.sql.Types</code>.  This method returns the
+     * primitive Class objects (e.g. Integer.TYPE) rather than
+     * the wrapper class objects (Integer.class). This means that
+     * you can't call a method which takes one of the
+     * <code>java.lang</code> wrapper types. Since primitive
+     * types are much more common as arguments, this doesn't seem
+     * like too much of a problem.
+     *   <p>
+     * The JDBC type to Class mappings are based on the JDBC type
+     * to Java type tables in the JDBC books:
+     * <ul>
+     * <li>Table 21.6.1 in <i>JDBC Data Access with Java</i>,
+     *      Hamilton, Cattell, Fisher
+     * <li>Table 47.9.1 in <i>JDBC API Tutorial and Reference,
+     *      Second Edition</i>, White, Fisher, Cattell, Hamilton,
+     *      Hapner
+     * </ul>
+     *
+     * @param aJdbcType a JDBC type code from <code>java.sql.Types</code>
+     * @return the class object corresponding to the given type
+     */
+    private Class mGetClass(int aJdbcType)
+    {
+        switch (aJdbcType) 
+        {
+        case Types.BINARY:
+        case Types.VARBINARY:
+        case Types.LONGVARBINARY:
+            byte[] b = new byte[0];
+            return b.getClass();
+
+        case Types.TINYINT:             return Byte.TYPE;
+        case Types.SMALLINT:            return Short.TYPE;
+        case Types.INTEGER:             return Integer.TYPE;
+        case Types.BIGINT:              return Long.TYPE;
+        case Types.REAL:                return Float.TYPE;
+        case Types.FLOAT: 
+        case Types.DOUBLE:              return Double.TYPE;
+        case Types.DECIMAL: 
+        case Types.NUMERIC:             return BigDecimal.class;
+        case Types.BIT:                 return Boolean.TYPE;
+        case Types.CHAR: 
+        case Types.VARCHAR:
+        case Types.LONGVARCHAR:         return String.class;
+        case Types.DATE:                return java.sql.Date.class;
+        case Types.TIME:                return java.sql.Time.class;
+        case Types.TIMESTAMP:           return java.sql.Timestamp.class;
+        case Types.OTHER: 
+        default: 
+            return String.class;  // String will be the actual type.
+        }
+    }
+
     /**
      * Retrieves a StatementHolder from the statement table.
      *
      * @param aStatementHandle the statement handle to retrieve
      * @return the corresponding StatementHolder from the cache
      * @exception Exception if the statement handle does not correspond
-     *      to an entry in the statement table
-     */
+     *      to an entry in the statement table */
     private StatementHolder mGetStatementHolder(int aStatementHandle)
         throws DbdException
     {
         StatementHolder holder = (StatementHolder) mStatementTable.get(
             new Integer(aStatementHandle));
-        // ??? Should this be a runtime exception? Can the user invoke this?
+        // ??? Should this be a runtime exception? Can the user cause this?
         if (holder == null)
             throw new DbdException(DbdException.gINVALID_STATEMENT_HANDLE);
         return holder;
@@ -1216,3 +1597,4 @@ class StatementHolder
         return mStatementProperties;
     }
 }
+
